// BEACHTE: Dieses Programm benutzt 64 bit Integer, um die Bit-Darstellung von double-Variablen zu finden.
//          Es muss daher mit 64 bit kompiliert werden, wenn die Bit-Darstellung einer 64 bit Variable
//          (int oder float) bestimmt werden soll.

#include <stdio.h>
#include <stdlib.h> // fuer exit()
#include <string.h> // fuer strcmp()

#define STR_LENGTH 255

int main(void)
{
  int i;
  int bits64 = 0; // 1: 64 bit, 0: 32 bit

  char data_type[STR_LENGTH];
  char bit_pattern[64]; // enthaelt am Ende die gesuchte Bit-Darstellung der Zahl

  // vier verschiedene Datentypen: 32 bit int (d), 64 bit int (ld), 32 bit float (f), 64 bit float (lf)

  int      x_d;
  long int x_ld;
  float    x_f;
  double   x_lf;

  // unsigned int Variablen passender Groesse, in die das Bit-Muster zum Auslesen kopiert wird

  unsigned int      buffer_u;
  unsigned long int buffer_lu;
 
  // frage Datentyp ab und lese die Zahl ein

  printf("\nDatentyp?\n(d = 32 bit int, ld = 64 bit int, f = 32 bit float, lf = 64 bit float)\n\n");
  scanf("%s", data_type);
  printf("\nZahl?\n\n");
  
  // strcmp(x,y) = 0 -> x = y

  if(strcmp(data_type, "d") == 0) // 32 bit int
    scanf("%d", &x_d);
  else if(strcmp(data_type, "ld") == 0) // 64 bit int
    {
      scanf("%ld", &x_ld);
      bits64 = 1; // flag setzen
    }
  else if(strcmp(data_type, "f") == 0) // 32 bit float
    scanf("%f", &x_f);
  else if(strcmp(data_type, "lf") == 0) // 64 bit float
    {
      scanf("%lf", &x_lf);
      bits64 = 1; // flag setzen
    }
  else // Fehler! 
    {
      printf("Fehler: unbekannter Datentyp!\n\n");
      exit(1); // beendet das Programm
    }
  printf("\n");

  // Idee zur Bestimmung der Bit-Darstellung:
  // Wir wollen das Bit-Muster der (float bzw. negativen int) Variable ohne Veraenderung (!) 
  // auf eine unsigned int Variable uebertragen, um es dann aus dieser mit bitweisen Operationen
  // auslesen zu koennen. Dazu: 
  // - bestimme mit "&" die Position im Speicher, wo der Wert der Variable in einem 32 oder 64 bit Block 
  //   abgelegt ist (Speicheradresse bzw. Pointer)
  // - wandle diesen Pointer manuell zu einem unsigned int Pointer um (type casting),
  //   d.h. suggeriere dem Programm, dass an der Speicherstelle tatsaechlich eine Integer-Zahl liegt,
  //   auch wenn dies vielleicht gar nicht der Fall ist
  // - greife mit "*" auf den Speicherinhalt zu und kopiere diesen (d.h. die Bits) in einen unsigned Integer, 
  //   dessen Groesse zur Laenge des Speicherblocks passen muss

  if(strcmp(data_type, "d") == 0) // 32 bit int
    buffer_u = *((unsigned int *) &x_d);
  else if(strcmp(data_type, "ld") == 0) // 64 bit int
    buffer_lu = *((unsigned long int *) &x_ld);
  else if(strcmp(data_type, "f") == 0) // 32 bit float
    buffer_u = *((unsigned int *) &x_f);
  else // 64 bit float
    buffer_lu = *((unsigned long int *) &x_lf);

  // Bestimme jetzt die Bit-Darstellung der Buffer- (bzw. der urspruenglichen) Variable mit Hilfe von bitweisen
  // Operationen. Dazu kann man die beiden folgenden Operatoren verwenden:
  //  
  // 1) >>
  //    "x >> i" (x und i Integer) hat als Ergebnis einen Integer, in dem das urspruengliche Bit-Muster
  //    von x um i Bits nach rechts verschoben ist. Bits, die rechts "herausgeschoben" werden, fallen dabei weg
  //    und Bits, die links frei werden, werden auf 0 gesetzt.
  //  
  // 2) & (bitwise AND)
  //    "x & y" vergleicht die Integer x und y Bit fuer Bit. Im Ergebnis-Integer ist das i. Bit genau dann gleich 1,
  //    wenn das entsprechende Bit sowohl in x als auch in y gleich 1 ist.

  // Zum Auslesen des i. Bits aus der Buffer-Variable wird dieses Bit (mit dem Rest des Bit-Musters) ganz nach rechts
  // geschoben. Anschliessend wird mittels "&" mit dem Bit-Muster von 1 verglichen (000...000001). Das Ergebnis ist
  // dann ein Integer, bei dem alle Bits ausser dem ganz rechts auf 0 gesetzt sind. Das Bit ganz rechts behaelt den
  // vorigen Wert bei, und da alle anderen Bits gleich 0 sind, entspricht der Wert des entstehenden Integers gerade
  // dem gesuchten Wert des i. Bits (0 oder 1).

  if(bits64 == 1) // 64 bit Variable
    for(i=63; i>=0; i--)
      bit_pattern[63-i] = (buffer_lu >> i) & 1;
  else // 32 bit Variable
    for(i=31; i>=0; i--)
      bit_pattern[31-i] = (buffer_u >> i) & 1;

  // Ausgabe der gefundenen Bit-Darstellung

  printf("Bit-Darstellung ");
  if(bits64 == 1)
    printf("(64 bit):\n\n");
  else
    printf("(32 bit):\n\n");
  
  for(i=0; i<32; i++)
    printf("%d ", bit_pattern[i]);
  printf("\n");
  if(bits64 == 1)
    for(i=32; i<64; i++)
      printf("%d ", bit_pattern[i]);
  
  printf("\n\n");

  return 0;
}
